function LootSquid:AddEvent(event, func)
    self.eventHandlers[event] = func
end

function LootSquid:RegisterEvents()

    self:LogDebug("RegisterEvents()")

    -- hook chat events
    --self.eventHandlers["CHAT_MSG_BN_WHISPER"] = self.OnChatMessage
    self.eventHandlers["CHAT_MSG_RAID"] = self.OnChatMessage
    self.eventHandlers["CHAT_MSG_WHISPER"] = self.OnChatMessage
    self.eventHandlers["LOOT_OPENED"] = self.OnLootOpened
    self.eventHandlers["LOOT_CLOSED"] = self.OnLootClosed
    self.eventHandlers["LOOT_SLOT_CHANGED"] = self.OnLootSlotChanged
    self.eventHandlers["LOOT_SLOT_CLEARED"] = self.OnLootSlotCleared
    self.eventHandlers["GET_ITEM_INFO_RECEIVED"] = self.OnGetItemInfoReceived

    -- register all event handlers on the main frame
    for k in pairs(self.eventHandlers) do
        self.frames.mainFrame:RegisterEvent(k)
    end

    -- add on OnEvent handler on the main frame
    self.frames.mainFrame:SetScript("OnEvent",
        function(mainFrame, event, ...)
        --self:LogInfo("OnEvent: " .. event)
            self.eventHandlers[event](self, ...)
        end)

    self.frames.mainFrame:SetScript("OnUpdate", LootSquid.OnUpdate)
end

function LootSquid:OnUpdate()

    if not LootSquid:IsActive() then
        return
    end

    --[[
    lastChatRequestEventTime = 0,
    lastGetItemInfoEventTime = 0,
    lastRequestFrameUpdatedTime = 0,
    -- ]]

    --print(tostring(LootSquid.lastChatRequestEventTime) .. " vs " .. tostring(LootSquid.lastRequestFrameUpdatedTime))

    local now = GetTime()

    if LootSquid.lastChatRequestEventTime > LootSquid.lastRequestFrameUpdatedTime
        or LootSquid.lastGetItemInfoEventTime > LootSquid.lastRequestFrameUpdatedTime then

        -- request update
        LootSquid:LogDebug("Requesting update")
        LootSquid.lastRequestFrameUpdateRequestTime = now
        LootSquid.lastChatRequestEventTime = 0
        LootSquid.lastGetItemInfoEventTime = 0
    else
        --LootSquid:LogDebug("not request for update")
    end

    local triggerDelaySecs = 1.0
    local triggerWindowSecs = 5.0

    if LootSquid.lastRequestFrameUpdateRequestTime
            and (now - LootSquid.lastRequestFrameUpdateRequestTime) > triggerDelaySecs
            and (now - LootSquid.lastRequestFrameUpdateRequestTime) < (triggerDelaySecs + triggerWindowSecs) then

        LootSquid:LogDebug("Updating request frame because 3 seconds have passed.")

        LootSquid.lastRequestFrameUpdateRequestTime = 0
        LootSquid.lastRequestFrameUpdatedTime = now
        LootSquid:UpdateRequestFrame()
    end
end

function LootSquid:IsPlayerMasterLooter()
    local master, partyMaster, raidMaster = GetLootMethod()
    local playerName = UnitName("player")

    local raidName = "raid" .. tostring(raidMaster)

    if master and raidMaster and playerName == UnitName(raidName) then
        self:LogDebug("IsPlayerMasterLooter: " .. tostring(true))
        return true
    else
        self:LogDebug("IsPlayerMasterLooter: " .. tostring(false))
        return false
    end
end

function LootSquid:OnGetItemInfoReceived(a, b, c, d, e, f, g, h)

    if not LootSquid:IsActive() then
        return
    end

    LootSquid:LogDebug("OnGetItemInfoReceived: "
            .. tostring(a) .. ", " .. tostring(b) .. ", " .. tostring(c) .. ", " .. tostring(d) .. ", "
            .. tostring(e) .. ", " .. tostring(f) .. ", " .. tostring(g) .. ", " .. tostring(h) .. ", ")

    LootSquid.lastGetItemInfoEventTime = GetTime()
end

function LootSquid:OnLootSlotChanged(a, b, c, d, e, f, g, h)
    self:LogDebug("OnLootSlotChanged()")
    --    LootSquid:LogInfo("OnLootSlotChanged: "
    --            ..tostring(a)..", "..tostring(b)..", "..tostring(c)..", "..tostring(d)..", "
    --            ..tostring(e)..", "..tostring(f)..", "..tostring(g)..", "..tostring(h)..", ")
end

--[[
-- Fires when the contents of a loot slot are removed.
-- Can fire due to the player looting the slot's contents
-- or due to them being taken by another group member.
 ]]
function LootSquid:OnLootSlotCleared(event, slotId)

    if not LootSquid:IsActive() then
        return
    end

    self:LogDebug("OnLootSlotCleared. slotId: " .. tostring(slotId))
    -- slotId: The numeric id of the slot that was looted.
    LootSquid:UpdateLootTable(false)
end

function LootSquid:OnChatMessage(message, sender, language, channelString, target, flags,
unknown, channelNumber, channelName, unknown, counter, guid)

    if not LootSquid:IsActive() then
        return
    end

    if LootSquid.debug.bypassSafeCalls then
        LootSquid:HandleOnChatMessage(message, sender)
    else
        local success, error = pcall(LootSquid.HandleOnChatMessage, nil, message, sender)
        if not success then
            LootSquid:LogError("OnChatMessage failed: " .. tostring(error))
        end
    end
end

function LootSquid:HandleOnChatMessage(message, sender)

    if not message or not sender then
        return
    end

    LootSquid:LogDebug("OnChatMessage. Sender: " .. tostring(sender) .. ", message: " .. tostring(message))

    -- do not process chat message unless master looter
    if not (LootSquid.debug.alwaysParseChat or LootSquid:IsPlayerMasterLooter()) then
        return
    end

    local a, b, Color, Ltype, Id, Enchant, Gem1, Gem2, Gem3, Gem4, Suffix, Unique, LinkLvl, Name =
    string.find(message,
        "|?c?f?f?(%x*)|?H?([^:]*):?(%d+):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%-?%d*):?(%-?%d*):?(%d*):?(%d*):?(%-?%d*)|?h?%[?([^%[%]]*)%]?|?h?|?r?")

    local isNeed = LootSquid:strStartsWith(string.lower(message), "!need")
    local isGreed = LootSquid:strStartsWith(string.lower(message), "!greed")
    local isPass = LootSquid:strStartsWith(string.lower(message), "!pass")

    if (_G.LootSquid.session.currentLootSlot < 1 and (isNeed or isGreed or isPass)) then
        LootSquid:WhisperToPlayer(sender, "The loot master does not currently have an active loot session")
        return
    end

    if (isNeed) then

        _G.LootSquid.session.playersNeeding[sender] = { isNeeding = true, currentItemId = Id }
        _G.LootSquid.session.playersGreeding[sender] = nil
        LootSquid:WhisperToPlayer(sender, "Your need request have been received. Please wait.")
    end

    if (isGreed) then

        _G.LootSquid.session.playersGreeding[sender] = { isGreeding = true, currentItemId = Id }
        _G.LootSquid.session.playersNeeding[sender] = nil
        LootSquid:WhisperToPlayer(sender, "Your greed request have been received. Please wait.")
    end

    if (isPass) then

        _G.LootSquid.session.playersGreeding[sender] = nil
        _G.LootSquid.session.playersNeeding[sender] = nil
        LootSquid:WhisperToPlayer(sender, "Your pass request have been received. You have been removed from concideration.")
    end

    LootSquid.lastChatRequestEventTime = GetTime()
end

function LootSquid:UpdateLootTable(doOpenGui)

    -- if player is the master looter
    if (_G.LootSquid.debug.alwaysParseLoot == true
            or (LootSquid:IsPlayerMasterLooter())) then

        for k in pairs(_G.LootSquid.session.loot) do
            _G.LootSquid.session.loot[k] = nil
        end

        local lootThreshold = GetLootThreshold()

        local lootSlots = GetNumLootItems()
        local lootSlotsAtleastThreshold = 0

        _G.LootSquid.session.lootSlots = lootSlots

        for lootSlot = 1, lootSlots do

            --[[
            -- The lootSlot will remain concistent for items as long as the loot box remains open.
            -- If the loot box is closed after some items have been looted, lootSlot will have changed
            -- for items which originally was placed after a looted item.
            -- ]]
            local itemTexture, itemName, itemQuantity, itemQuality, itemLocked = GetLootSlotInfo(lootSlot)

            if itemQuality and itemQuality >= lootThreshold then
                lootSlotsAtleastThreshold = lootSlotsAtleastThreshold + 1
            end

            local itemLootType = GetLootSlotType(lootSlot)


            --            self:LogInfo("GetLootSlotInfo "
            --                    .. tostring(lootSlot) .. ", "
            --                    .. tostring(itemName) .. ", "
            --                    .. tostring(itemQuantity) .. ", "
            --                    .. tostring(itemQuality) .. ", "
            --                    .. tostring(itemLocked) .. ", "
            --                    .. tostring(itemLootType)
            --            )

            local itemLink = GetLootSlotLink(lootSlot)

            _G.LootSquid.session.loot[lootSlot] = {
                itemName = itemName,
                itemQuantity = itemQuantity,
                itemQuality = itemQuality,
                itemLocked = itemLocked,
                itemLootType = itemLootType,
                itemLink = itemLink,
                candidates = {}
            }

            for candidateIndex = 1, 40 do

                local candidateName = GetMasterLootCandidate(lootSlot, candidateIndex)
                if candidateName then

                    _G.LootSquid.session.loot[lootSlot].candidates[candidateName] = {
                        candidateIndex = candidateIndex,
                    }

                else
                    LootSquid:LogDebug("no candidate name")
                end
            end
        end

        if ((_G.LootSquid.debug.alwaysParseLoot and doOpenGui)
                or
                (lootSlotsAtleastThreshold > 0
                        and doOpenGui
                        and lootSlots > 0
                )) then
            self:OpenLootGui()
        else
            self:OnUpdateLootTable()
        end
    end
end

function LootSquid:OnLootOpened(autoLoot)
    LootSquid:LogDebug("OnLootOpened. autoLoot: " .. tostring(autoLoot))
    LootSquid:UpdateLootTable(LootSquid:IsPlayerMasterLooter())
end

function LootSquid:OnLootClosed()
    if LootSquid.frames.window and LootSquid.frames.window:IsVisible() then
       LootSquid:LogError("Loot window closed before loot squid window. Please do not move while processing loot.")
       LootSquid.frames.window:Hide()
    end
end

LootSquid.components.events = true